"""
TagCash.

Usage:
  tagcash [-t <TAGS>] [-a] [<FILE>... | -]

Arguments:
  FILE  One or more transaction files
  -     Use stdin (default)

Options
  -t TAGS --tags=TAGS  Comma-separated tags to filter, no spaces.
                       Defaults to use all tags.
  -a --all             Show an extra table with all the tags.
"""
import fileinput
from collections import defaultdict

from docopt import docopt
from terminaltables import SingleTable

from .entry import parse_lines, update_balance


def parse_entries(tags, files):
    """Parse all lines of all files that have any tag of ``tags``."""
    entries_by_tag = defaultdict(list)
    lines = fileinput.input(files)
    for entry in parse_lines(lines, tags):
        entries_by_tag[entry.tag].append(entry)
    return entries_by_tag


def print_tag_table(title, entries):
    """Show entries as a table."""
    header = ('Date', 'Amount', 'Balance', 'Description')
    rows = [(entry.date,
             '{:,.2f}'.format(entry.amount),
             '{:,.2f}'.format(entry.balance),
             entry.description) for entry in entries]
    table_data = [header] + rows
    table = SingleTable(table_data, title=title)
    table.justify_columns[1] = 'right'
    table.justify_columns[2] = 'right'
    print(table.table)


def main():
    """Entry point."""
    args = docopt(__doc__, version='1.0.0b2')
    if args['--tags'] is None:
        tags = None
    else:
        tags = set(tag.lower() for tag in args['--tags'].split(','))

    entries = parse_entries(tags, args['<FILE>'])
    for tag in sorted(entries):
        update_balance(entries[tag])
        print_tag_table(tag, entries[tag])

    if args['--all']:
        all_entries = [entry for tag_entries in entries.values()
                       for entry in tag_entries]
        update_balance(all_entries)
        print_tag_table('All Tags', all_entries)



"""Parse and manage transactions."""
import re
from sys import stderr


_LINE_RE = re.compile(r"""^
                      (?P<date>\d{4}-\d{2}-\d{2}) \s+
                      (?P<amount>-?[\d\.,]+)      \s+
                      (?P<description>.+)         \s+
                      (?P<tags>[-\w,]+)
                      $""", re.X)


class Entry:  # pylint: disable=too-few-public-methods
    """One event of one tag."""

    def __init__(self, tag, date, amount, description):
        """Set all instance variables."""
        self.date = date
        self.amount = self._parse_amount(amount)
        self.description = description
        self.balance = None
        # Negate value if tag is negative
        if tag[0] == '-':
            self.tag = tag[1:]
            self.amount *= -1
        else:
            self.tag = tag

    @staticmethod
    def _parse_amount(amount):
        try:
            return int(amount)
        except ValueError:
            pass  # let's try float
        try:
            return float(amount)
        except ValueError:
            pass  # let's try checking decimal point and thousand separator

        # digit + n digits, points or commas + dot or comma + 2 digits
        match = re.match(r'-?\d[\d\.,]+([\.,])\d\d', amount)
        if match:
            digits = re.sub(r'[\.,]', '', amount)
            return int(digits) / 100

        raise ValueError('Couldn\'t parse amount "{}"'.format(amount))


def update_balance(entries):
    """Sort entries by date and update balances."""
    entries.sort(key=lambda entry: entry.date)
    balance = 0
    for entry in entries:
        entry.balance = balance + entry.amount
        balance = entry.balance


def parse_lines(lines, wanted_tags):
    """Parse lines having specific tags or all if none specified."""
    for line in lines:
        match = _LINE_RE.match(line)
        if match:
            parsed_line = match.groupdict()
            yield from _create_entries(wanted_tags, parsed_line)
        else:
            print('Couldn\'t parse line "{}"'.format(line.rstrip()),
                  file=stderr)


def _create_entries(wanted_tags, parsed_line):
    """Yield entries having selected tags."""
    tags = parsed_line.pop('tags').split(',')
    for tag in tags:
        abs_tag = tag[1:].lower() if tag[0] == '-' else tag.lower()
        if not wanted_tags or abs_tag in wanted_tags:
            yield Entry(tag, **parsed_line)



"""A setuptools based setup module."""
from pathlib import Path

from setuptools import setup

HERE = Path().parent.resolve()  # pylint: disable=no-member

# Get the long description from the README file
README = HERE / 'README.rst'
with open(str(README), encoding='utf-8') as f:
    LONG_DESC = f.read()

setup(
    name='tagcash',
    version='1.0.0b2',
    description='Finances with tags in CLI',
    long_description=LONG_DESC,
    url='https://github.com/cemsbr/tagcash',
    author='Carlos Eduardo Moreira dos Santos',
    author_email='cems@cemshost.com.br',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'Topic :: Office/Business :: Financial :: Accounting',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='accounting finance finances cli terminal',
    packages=['tagcash'],
    install_requires=[
        'docopt',
        'terminaltables',
    ],
    extras_require={
        'dev': [
            # coverage version:
            # - incomplete-file-path-in-xml-report
            #   https://bitbucket.org/ned/coveragepy/issues/578/
            # - Code climate requires < 4.4
            'coverage<4.4',
            'eradicate',
            'pip-tools',
            'rstcheck',
            'tox',
            # Python 3.5 support since 1.4.0
            'yala>=1.4.0',
        ]
    },
    entry_points={
        'console_scripts': [
            'tagcash=tagcash.interface:main',
        ],
    },
    test_suite='tests',
)
